@@ -32,6 +32,8 @@
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+
+#define BIND_8_COMPAT
 #include <arpa/nameser.h>
 #include <netdb.h>
 #include <resolv.h>
@@ -48,14 +50,25 @@
 #include "dns.h"
 #include "net.h"
 
+/* OSX  Needs this. I don't know how to enable this for them automatically. 
+ * Should be easy with autoconf. Please submit a patch if you know 
+ * autoconf.... -- REW
+ */
+#if 0 
+#include "nameser8_compat.h"
+#endif
+
+
 #ifdef NO_STRERROR
 extern int sys_nerr;
 extern char *sys_errlist[];
 #define strerror(errno) (((errno) >= 0 && (errno) < sys_nerr) ? sys_errlist[errno] : "unlisted error")
 #endif
 
+#if !HAVE_DECL_ERRNO
 /*  Hmm, it seems Irix requires this  */
 extern int errno;
+#endif
 
 extern int af;
 
@@ -160,12 +173,15 @@ char *rrtypes[] = {
    "Resource reference",
 };
 
+
+/* Please don't use a trailing comma in enumerations: It doesn't
+   work on all compilers -- REW */
 enum {
    RR_UNKNOWN,
    RR_QUERY,
    RR_ANSWER,
    RR_AUTHORITY,
-   RR_RESOURCE,
+   RR_RESOURCE
 };
 
 typedef struct {
@@ -230,7 +246,7 @@ enum {
    STATE_FAILED,
    STATE_PTRREQ1,
    STATE_PTRREQ2,
-   STATE_PTRREQ3,
+   STATE_PTRREQ3
 };
 
 #define Is_PTR(x) ((x->state == STATE_PTRREQ1) || (x->state == STATE_PTRREQ2) || (x->state == STATE_PTRREQ3))
@@ -447,7 +463,7 @@ int longipstr( char *s, ip_t *dst, int af )
 }
 
 
-struct hostent * dns_forward(char *name)
+struct hostent * dns_forward(const char *name)
 {
   struct hostent *host;
 
@@ -1357,7 +1373,7 @@ char *dns_lookup(ip_t * ip)
 #ifdef ENABLE_IPV6
 /* Returns an ip6.arpa character string. */
 void addr2ip6arpa( ip_t * ip, char * buf ) {
-  char * p = (char *) ip;
+  unsigned char * p = (unsigned char *) ip;
   char * b = buf;
   int i;
 
